/****
 * CSignKeyWndGrid.c
 *
 *	Sign key window methods for the SignWriter key editbor.
 *
 *  Copyright 1996 The Center for Sutton Movement Writing. All rights reserved.
 *
 ****/

 
/**
 *
 *  This class subclasses TCL's CGridSelector class to implement
 *  the key editor's sign key window.
 *
 **/


#include "CSignKeyWndGrid.h"
#include "CBureaucrat.h"
#include "CFileWndKeys.h"
#include "CSelector.h"
#include "CSignKeyWndTask.h"
#include "CWindow.h"
#include "Events.h"
#include "LongCoordinates.h"
#include "LongQD.h"
#include "KeyEditConsts.h"
#include "Keys.h"
#include "Symbols.h"
#include "TBUtilities.h"





/*** Global Variables ***/

extern void			*gClipRec;			/* Clipboard */
extern ObjType	 	 gClipType;

extern CFileWndKeys	*gKeyPanes;			/* File window keyboard */





/****************************  Initialization  ****************************/


/***
 * ISignKeyWndGrid
 *
 ***/
 
void CSignKeyWndGrid::ISignKeyWndGrid (CView *anEnclosure, CBureaucrat *aSupervisor,
							   		   Byte keyCode)
{
	Box	  var;
	
	
	itsKeyCode  = keyCode;
		
	itsOn = FALSE;
	
	GetSignKeyVar(itsKeyCode, 0, &var);			// show cursor in first box			  
	SetSelectType(var, none);
	
	inherited::IGridSelector(anEnclosure, aSupervisor,
							 0, 0, sizFIXEDSTICKY, sizFIXEDSTICKY,
							 1, 0, SKGRID_ROWS, SKGRID_COLS,
							 SKCELL_WIDTH, SKCELL_HEIGHT);							 
						 			
}




/******************************  Grid mapping  *******************************/


/***
 * ItemToKeyVar
 *
 *	Convert a grid class item number
 *  to the corresponding 0-based key 
 *  variation.
 *
 ***/
 
short CSignKeyWndGrid::ItemToKeyVar (short item)

{	
	if (item >= 9) item--;		// righthand 4 keys
	if (item >= 6) item--;		// middle 2 keys
	
	return (item - 1);			// vars 0-based
	
}
 
 
 
/***
 * KeyVarToItem
 *
 *	Convert a key variation to the
 *  corresponding grid class item.
 *
 ***/
 
short CSignKeyWndGrid::KeyVarToItem (short keyVar)

{
	if (keyVar >= 6) keyVar++;		// righthand 4 keys
	if (keyVar >= 4) keyVar++;		// middle 2 keys
	
	return (keyVar + 1);			// items 1-based
	
}


 
/***
 * ValidItem
 *
 *  Return TRUE if the specified item
 *  is a defined key cell.
 *
 ***/
 
Boolean CSignKeyWndGrid::ValidItem (short item) 

{
 	return !(item == 5 || item == 8);	// skip unused cells
}
 	
 	
 	
 	
/********************************  Drawing  ********************************/


/***
 * DrawItem (override)
 *
 *  Draw the specified grid item.  The item numbering
 *  scheme is as defined for CGridSelector.
 *
 ***/
 
void CSignKeyWndGrid::DrawItem (short theItem, Rect *theBox)
 
{ 
 	if (!ValidItem(theItem)) {
 		FillRect(theBox, ltGray);
 		return;
 	}
	
	Prepare();
	
	EraseRect(theBox);
	DrawSignKey(itsKeyCode, ItemToKeyVar(theItem), &thePort->portBits, theBox, FALSE);
	
}



/***
 * UpdateKeyVar
 *
 *  Update the specified key
 *  variation after key editing.
 *
 ***/

void CSignKeyWndGrid::UpdateKeyVar (short keyVar)
 
{
	short 	 item = KeyVarToItem(keyVar);
	Rect	 keyBox;
	Box		 keyRec;
	ObjType  selType;

	
	FindItemBox(item, &keyBox);
	DrawItem(item, &keyBox);							// redraw key box
	
	
	GetSignKeyVar(itsKeyCode, keyVar, &keyRec);	
	GetSelectType(keyRec, &selType);
	
	if (selType == box)									// redo box select
		ToggleBox(item);		
		
	else												// blink is off
		itsOn = FALSE;
	
}



/***
 * ToggleBox
 *
 *  Invert key box.
 *
 ***/

void CSignKeyWndGrid::ToggleBox (short item)
 
{	
	Rect  keyBox;
	
	
	FindItemBox(item, &keyBox);
	
	Prepare();
	InvertRect(&keyBox);
			
}



/***
 * ToggleCursor
 *
 *  Draw insertion point in 
 *  the specified grid item.
 *
 ***/
 
void CSignKeyWndGrid::ToggleCursor (short item)

{
 	Rect  keyBox;

	FindItemBox(item, &keyBox);
	
	Prepare();
	PenMode(patXor);

	MoveTo(keyBox.left + KCURS_LEFT, keyBox.top + KCURS_TOP);
	Line(0, KCURS_LEN);
	
	itsOn = !itsOn;
	
}

 
 
/***
 * ToggleSelect
 *
 *  Draw/erase selected symbol.
 *
 ***/
 
void CSignKeyWndGrid::ToggleSelect (short item)

{
	Rect   keyBox;
	short  keyVar = ItemToKeyVar(item);

	
	FindItemBox(item, &keyBox);

	Prepare();
	
	if (!itsOn)
		DrawSignSelect(itsKeyCode, keyVar, &thePort->portBits, &keyBox);
	else
	 	EraseSignSelect(&thePort->portBits);			
	 		
	itsOn = !itsOn;
	
}




/********************************  Selection  ********************************/


/***
 * BlinkSelect 
 *
 *  Blink the insertion point or
 *  currently selected symbol.
 *
 ***/
 
void CSignKeyWndGrid::BlinkSelect (long *maxSleep)

{	
	short    item = CSelector::GetSelection();
	short    keyVar = ItemToKeyVar(item);
	Box		 var;
	ObjType  selType;
	
	
	GetSignKeyVar(itsKeyCode, keyVar, &var);				  
	GetSelectType(var, &selType);	
	
	switch (selType) {

		case box:
		
			*maxSleep = GetCaretTime();
			
			// boxes don't blink
			
			break;
			
			
		case sym:
						
			*maxSleep = (itsOn) ? SYM_OFF : SYM_ON;

			ToggleSelect(item);
		
			break;
			
		
		case none:
			
			*maxSleep = GetCaretTime();
			
			ToggleCursor(item);
			
			break;
		
	}
			
}



/***
 * Deselect
 *
 *  Deselect the specified item and
 *  update the display accordingly.
 *
 ***/
 
void CSignKeyWndGrid::Deselect (short item)

{
	short    keyVar;
	Box		 var;
	ObjType  selType;


	if (!ValidItem(item)) return;
	
	
	keyVar = ItemToKeyVar(item);
	
	GetSignKeyVar(itsKeyCode, keyVar, &var);				  
	GetSelectType(var, &selType);
	
		
	switch (selType) {
	
		case box:
		
			ToggleBox(item);							// de-invert box
			break;
			

		case sym:
			
			if (!itsOn) ToggleSelect(item);				// draw selected sym		
			break;
			
			
		case none:

			if (itsOn) ToggleCursor(item);				// turn off cursor
			break;		
				
	}
	
	
	SetSelectType(var, none);							// do deselect
	
	itsOn = FALSE;										// blink is off

}



/***
 * HiliteItem (override default)
 *
 *  Replace HiliteItem with explicit
 *  grid selection (in DoClick).
 *
 ***/

void CSignKeyWndGrid::HiliteItem (short theItem, HiliteState state)

{
	// do nothing
}




/***
 * Deactivate (override default)
 *
 *  Deselect before leaving keybox window.
 *
 ***/
 
void CSignKeyWndGrid::Deactivate (void)
 
{
	Deselect(CSelector::GetSelection());
	
	inherited::Deactivate();					// do inherited
	
}



/***
 * Close
 *
 *  Deselect before closing keybox window.
 *
 ***/
 
void CSignKeyWndGrid::Close (void)
 
{
	Deselect(CSelector::GetSelection());
}

	

/***
 * SelectionType
 *
 *  Return the type of the object currently
 *  selected in the sign key window grid.
 *
 ***/
 
ObjType CSignKeyWndGrid::SelectionType (void)

{
	Box		 var;
	ObjType  selType;
	
	
	GetSignKeyVar(itsKeyCode, ItemToKeyVar(CSelector::GetSelection()), &var);				  
	GetSelectType(var, &selType);
	
	return selType;
	
}




/******************************  Edit commands  *****************************/

	
/***
/* 
/*  DoCopy
/*
/*  Copy selected item to the clipboard. Not undoable.
/*
 ***/
		
void CSignKeyWndGrid::DoCopy()

{
	short  keyVar = ItemToKeyVar(CSelector::GetSelection());
	Box	   rec;
	
		
	if (gClipRec != NULL)
		DisposKey(gClipType, gClipRec);
		
	GetSignKeyVar(itsKeyCode, keyVar, &rec);	
	CopySignKeyVar(rec, (Box *) &gClipRec);

	GetSelectType(rec, &gClipType);		
	// CSignKeyWindow::UpdateMenu ensures this is 'box' or 'sym'		
	
}
	


/***
/* 
/*  DoClear
/*
/*  Erase the currently selected item
/*
 ***/

void CSignKeyWndGrid::DoClear (short taskIndex)

{
	short			  keyVar = ItemToKeyVar(CSelector::GetSelection());
	CSignKeyWndClear *theClearTask;

						
	theClearTask = new CSignKeyWndClear;
	theClearTask->ISignKeyWndClear(taskIndex, this, itsKeyCode, keyVar);
	theClearTask->Do();
	
	itsSupervisor->Notify(theClearTask);
	itsLastTask = theClearTask;
	
}



/***
/* 
/*  DoPaste
/*
/*  Paste key record from the clipboard.
/*
 ***/

void CSignKeyWndGrid::DoPaste()

{
	short			  keyVar = ItemToKeyVar(CSelector::GetSelection());
	CSignKeyWndPaste *thePasteTask;

						
	thePasteTask = new CSignKeyWndPaste;
	thePasteTask->ISignKeyWndPaste(this, itsKeyCode, keyVar);
	thePasteTask->Do();

	itsSupervisor->Notify(thePasteTask);
	itsLastTask = thePasteTask;
	
}	


	
/***
/* 
/*  DoAdd
/*
/*  Add new symbol from the symbol window.
/*
 ***/
		
void CSignKeyWndGrid::DoAdd ()

{
	short			keyVar = ItemToKeyVar(CSelector::GetSelection());
	CSignKeyWndAdd *theAddTask;

						
	theAddTask = new CSignKeyWndAdd;
	theAddTask->ISignKeyWndAdd(this, itsKeyCode, keyVar);
	theAddTask->Do();

	itsSupervisor->Notify(theAddTask);
	itsLastTask = theAddTask;
	
}
	
	


/*****************************  Symbol commands  *****************************/



// Symbol state

static Byte	  *symBase;			// symbol base address
static short   symCode;			// symbol code 		
static short   symVar;			// sym variation
static short   symFill;			// fill variation
static short   symRot;			// rotation value
static short   symFlop;			// flop value
static short   symMark;			// mark value
static short   symXPos;			// X position
static short   symYPos;			// Y position



	
/***
/* 
/*  Setup
/*
/*  Do preliminary work for symbol modification methods.
/*
 ***/

void CSignKeyWndGrid::Setup (void)

{
	short	keyVar = ItemToKeyVar(CSelector::GetSelection());
	short	selInx;
	Box		key;
	
	
	GetSignKeyVar(itsKeyCode, keyVar, &key);			// get initial sym state				  	
	GetSelectSym(key, &selInx);
	GetSymBase(key, selInx, &symBase);

	GetSymInfo(symBase, &symCode, &symVar, &symFill, &symRot, &symFlop, 
									 	  &symMark, &symXPos, &symYPos);
			 
}



/***
/* 
/*  Update
/*
/*  Update modified symbol. The symbol state
/*  has been changed by one of the symbol
/*  command methods listed below.  (Dawdle 
/*  method redraws the modified symbol.)
/*
 ***/

void CSignKeyWndGrid::Update (void)

{
	LongPt	framePt, junk;

								 
	GetCoords(symCode, symVar, symFill, symRot, symFlop, 		// new symbol state
							&framePt, &junk, FALSE); 	
							
	symXPos = (SKCELL_WIDTH  - framePt.h) / 2;					// center sym in box				
	symYPos = (SKCELL_HEIGHT - framePt.v) / 2;						
																	
	SetSymInfo(symBase, symCode, symVar, symFill, symRot, 		// update sym code	
					symFlop, symMark, symXPos, symYPos);
								 								
}


	
/***
/* 
/*  DoChange
/*
/*  Change the symbol variation.
/*
 ***/

void CSignKeyWndGrid::DoChange (void)

{
	short  maxVar;
	
	
	Setup();
	  
	GetSymVars(symCode, &maxVar);		 
	
	if (KeyIsDown(SHFT_KEY)) {
	
		symVar--;
		if (symVar < 0)
			symVar = maxVar;
			
	} else {
	
		symVar++;
		if (symVar > maxVar)
			symVar = 0;
	}
	
	Update();
	
}



/***
/* 
/*  DoFill
/*
/*  Change the fill variation.
/*
 ***/
 
void CSignKeyWndGrid::DoFill (void)

{
	short  maxFill;
	
	
	Setup();
	  
	GetFillVars(symCode, symVar, &maxFill);		 
	
	if (KeyIsDown(SHFT_KEY)) {
	
		symFill--;
		if (symFill < 0)
			symFill = maxFill;
			
	} else {
	
		symFill++;
		if (symFill > maxFill)
			symFill = 0;
	}
	
	Update();
	
}
 	
 	
 		
/***
/* 
/*  DoMirror
/*
/*  Change symbol to its mirror image.
/*
 ***/
 
void CSignKeyWndGrid::DoMirror (void)

{
	ImageClass	imgClass;
	short		numRots;
	short  		maxRot, maxFlop;
	
	
	Setup();
	  
	GetAttrs(symCode, symVar, symFill, &imgClass, &maxRot, &maxFlop);

	numRots = maxRot + 1;
			
	if (numRots > 3) {
	
		if (imgClass == class3B)
			symRot += 4;
			
		// magic mirror formula - it works!
		symRot = (numRots - (symRot % numRots)) % numRots;
	}
		
	if (maxFlop == 1)
		symFlop = (symFlop + 1) % 2;
	
	Update();
	
}



/***
/* 
/*  DoRotate
/*
/*  Change the symbol rotation.
/*
 ***/
 
void CSignKeyWndGrid::DoRotate (void)

{
	ImageClass	junk1;
	short  		maxRot, junk2;
	
	
	Setup();
	  
	GetAttrs(symCode, symVar, symFill, &junk1, &maxRot, &junk2);

	if (!KeyIsDown(SHFT_KEY)) {
	
		symRot++;
		if (symRot > maxRot)
			symRot = 0;
			
	} else {
	
		symRot--;
		if (symRot < 0)
			symRot = maxRot;
	}
			
	Update();
			
}	
 		


/***
/* 
/*  DoCursor
/*
/*  Deselect symbol and redisplay cursor.
/*
 ***/

void CSignKeyWndGrid::DoCursor (void)

{
	short	keyVar = ItemToKeyVar(CSelector::GetSelection());
	Box		key;
	
	
	GetSignKeyVar(itsKeyCode, keyVar, &key);				  
	SetSelectType(key, none);

	UpdateKeyVar(keyVar);
	
}	



/***
 * DoSymCommand
 *
 *	Perform symbol edit command that corresponds
 *  to the specified key code.
 *
 ***/
 
void CSignKeyWndGrid::DoSymCommand (Byte keyCode)
 
{
	if (SelectionType() == sym)				// sym must be selected
	
	 	switch (keyCode) {
	 	
	 		case KEY_FILL:
	 		
	 			DoFill(); 			
	 			break;
	 		
	 		
	 		case KEY_CURSOR:
	 		
	 			DoCursor(); 			
	 			break;
	 		
	 		
	 		case KEY_MIRROR:
	 		
	 			DoMirror(); 			
	 			break;
	 			
	 	
	 		case KEY_CHANGE:
	 		
	 			DoChange(); 			
	 			break;
	 			
	 			
	 		case KEY_ROTATE:
	 		
	 			DoRotate(); 			
	 			break;
	 		
		}
		
}




/**********************************  Mouse  ***********************************/
	
	

/***
 * DoClick (override default method)
 *
 *	First filter out unused grid cells.
 *  If hit point is on a symbol, select it.
 *  If hit point is on two or more symbols,
 *  select the smallest.  If mouse is dragged,
 *  move selected symbol accordingly.
 *
 ***/ 

void CSignKeyWndGrid::DoClick (Point hitPt, short modifierKeys, long when)

{
	CSignKeyWndDrag  *theDragTask;
	
	short	  item    = FindItem(hitPt);
	short	  oldItem = CSelector::GetSelection();
	
	short	  symCount, keyVar, junk;
	short     symInx, selInx, selSym;
	short	  symSize, minSize;
	LongPt    framePt, notUsed;
	LongPt	  trackStart, trackEnd;
	Point	  keyPt;
	Rect      symFrame, boxFrame;
	LongRect  lboxFrame;
	ObjType	  selType;
	Byte	  *symBase;
	Box		  key;
	

	if (!ValidItem(item))									// skip gray boxes
		return;

	
	Deselect(oldItem);										// zap old selection
	
	
	if (item != oldItem) {									// zap last undoable action
	
		itsSupervisor->Notify(NULL);
		itsLastTask = NULL;
		
	}


	keyVar = ItemToKeyVar(item);							// get symbol count
	GetSignKeyVar(itsKeyCode, keyVar, &key);
	
	GetSymCount(key, &symCount);	

	
	FindItemBox(item, &boxFrame);							// hit pt -> key coords	
	keyPt.h = hitPt.h - boxFrame.left;
	keyPt.v = hitPt.v - boxFrame.top;
	
	
	minSize = 10000;										// look for symbol hit...
	selInx  = -1;

	for (symInx = 0; symInx < symCount; symInx++) {	
	
	
		GetSymBase(key, symInx, &symBase);					// get sym image frame

		GetSymInfo(symBase, &symCode, &symVar, &symFill, &symRot, &symFlop, 
							&junk, &symXPos, &symYPos);
		
		GetCoords(symCode, symVar, symFill, symRot, symFlop, 
								&framePt, &notUsed, FALSE); 	
								
		SetRect(&symFrame, 0, 0, framePt.h, framePt.v);
		OffsetRect(&symFrame, symXPos, symYPos);
		
		
		if (PtInRect(keyPt, &symFrame)) {					// hit a symbol
		
			symSize = (symFrame.bottom - symFrame.top) *
					  (symFrame.right  - symFrame.left);
			
			if (symSize < minSize) {	
				minSize = symSize;
				selInx = symInx;
			}
				
		}
	}
	
	
	if (selInx >= 0) {										// symbol selected...
	
	
		SetSelectType(key, sym);							
		SetSelectSym(key, selInx);
		
		
		UpdateKeyVar(keyVar);								// erase symbol
		
					
		QDToLongRect(&boxFrame, &lboxFrame);				// track mouse						
		QDToLongPt(hitPt, &trackStart);
	
		theDragTask = new CSignKeyWndDrag;
		theDragTask->ISignKeyWndDrag(this, item, key, selInx);

		ForceNextPrepare();
		Prepare();
		
		TrackMouse(theDragTask, &trackStart, &lboxFrame);
		
		theDragTask->Dispose();
		
	}


	inherited::DoClick(hitPt, modifierKeys, when);			// do inherited
	
}



/***
 * DoDoubleClick (override default method)
 *
 * 	Select the contents of the specified key box.
 *
 ***/
 
 void CSignKeyWndGrid::DoDoubleClick(void)
 
{
	short	item = CSelector::GetSelection();
	short	keyVar;
	Box		key;

	
	keyVar = ItemToKeyVar(item);
	GetSignKeyVar(itsKeyCode, keyVar, &key);				  	
	
	SetSelectType(key, box);								// select box	
						
	UpdateKeyVar(keyVar);										
											
}
